#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Parses information and generates a virulence and an antibiotic resistance profile using ARIBA results.
'''
## useful imports
import time
import io
import os
import re
import sys
import pandas as pd
from sys import argv
from io import open
from termcolor import colored

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts import ariba_caller
from BacterialTyper.scripts import card_trick_caller

#############################################################
def parse_vfdb(folder, sampleName, fileResults, fileFlags, summary, assembly_cutoff):
	"""Parses results from VFDB database.
	
	:param folder: ARIBA results generated using VFDB database
	:param sampleName: the sample name id
	:param fileResults: report.tsv file generated
	:param fileFlags: contains information for ``ariba expand flags`` (results_parser)
	:param summary: Summary file create by ``ariba summary`` (ariba_summary)
	:param assembly_cutoff: ARIBA assembly threshold cutoff [0-1].
	
	:type folder: string
	:type sampleName: string
	:type fileResults: string
	:type fileFlags: string
	:type summary: string
	:type assembly_cutoff: float
	 
	"""
	## 
	## Input is a folder for output results, sample name, and
	## fileResults is report.tsv generated by ariba
	## fileFlags was generated by results_parser calling ariba expand flags, 
	## summary is the summary file generated by ariba summary
	##

	## get data
	summary_data = pd.read_csv(summary, header=0, sep=',') 			## report_summary.csv :: parse information from ARIBA 
	fileFlags_data = pd.read_csv(fileFlags, header=0, sep='\t')		## flags_explain.tsv :: ariba expand flag: explained flags
	original_data = pd.read_csv(fileResults, header=0, sep='\t')	## report.tsv :: ariba report generated

	## summary data
	summary_data = summary_data.set_index('name')
	list_found_genes = summary_data.columns
	cluster_len = len(list_found_genes)

	# print info
	print ("\tCheck VFDB result: ", sampleName)
	print ("\t%s genes putatively involved in the strain virulence..." %cluster_len)

	############################################################################	
	## analyze each cluster confering virulence
	############################################################################	
	data = original_data.loc[original_data['cluster'].isin(list_found_genes)]
	colnames = ['Reference', 'ID', 'Protein-coding','Presence/Absence', 'Species', 'Variants', 'Description', 'Additional information']
	df_results = found_results(colnames, data, list_found_genes, 'VFDB')
	
	############################################################################	
	## analyze each cluster identified: might or might not confer virulence
	############################################################################	
	## get results: found, identified, partial
	df_identified = identified_results(original_data, "VFDB", list_found_genes, assembly_cutoff)

	##########################
	## generate excel sheet
	##########################
	name_excel = folder + '/' + sampleName + '_VFDB_results.xlsx'
	writer = pd.ExcelWriter(name_excel, engine='xlsxwriter') 	## open excel handle

	## write excel handle
	df_results.to_excel(writer, sheet_name='results')				## write results
	original_data.to_excel(writer, sheet_name='VFDB_ARIBA_report') 	## Original data from ARIBA
	fileFlags_data.to_excel(writer, sheet_name='flags') 			## ARIBA flags explained
	summary_data.to_excel(writer, sheet_name='ARIBA_summary') 		## ARIBA summary generated
	df_identified.to_excel(writer, sheet_name='identified') 		## Identified genes: ARIBA flags explained
	
	name_csv = folder + '/' + sampleName + '_VFDB_summary.csv'
	df_identified.to_csv(name_csv)

	## close excel handle		
	writer.save()											

	return (name_excel, name_csv)

#############################################################
def parse_card(folder, sampleName, fileResults, fileFlags, summary, assembly_cutoff, card_trick_info):
	## 
	## Parses results from CARD database.
	## Input is a folder for output results, sample name, and
	## fileResults is report.tsv generated by ariba
	## fileFlags was generated by results_parser calling ariba expand flags, 
	## summary is the summary file generated by ariba summary
	##
	
	## get data
	summary_data = pd.read_csv(summary, header=0, sep=',') 			## report_summary.csv :: parse information from ARIBA 
	fileFlags_data = pd.read_csv(fileFlags, header=0, sep='\t')		## flags_explain.tsv :: ariba expand flag: explained flags
	original_data = pd.read_csv(fileResults, header=0, sep='\t')	## report.tsv :: ariba report generated
	card_ontology = functions.get_data(card_trick_info + '/aro.obo.csv', ',', 'index_col=0') 		## read card_info generated for card_trick parse
	
	## summary data
	summary_data = summary_data.set_index('name')
	list_found_genes = summary_data.columns
	cluster_len = len(list_found_genes)
	
	## print info
	print ("\tCheck CARD result: ", sampleName)
	print ("\t%s genes putatively involved in resistance to some antibiotics..." %cluster_len)

	## subset
	data = original_data.loc[original_data['cluster'].isin(summary_data.columns)]
	
	############################################################################	
	## analyze each cluster confering resistance
	############################################################################	
	colnames = ['Reference', 'ID', 'Protein-coding','Presence/Absence', 'Variants', 'Description', 'Additional information']
	
	## get results: conferring resistance
	df_results = found_results(colnames, data, list_found_genes, 'CARD')

	## get results: found, identified, partial
	df_identified = identified_results(original_data, "CARD", list_found_genes, assembly_cutoff)

	############################################################################
	## use card-trick python package to get ontology for each term
	AROS_identified = list(df_identified['ID'])
	information_ontology = card_trick_caller.get_info_CARD(AROS_identified, 'ARO', card_ontology)

	##########################
	## generate excel sheet
	##########################

	## open excel handle
	name_excel = folder + '/' + sampleName + '_CARD_results.xlsx'
	writer = pd.ExcelWriter(name_excel, engine='xlsxwriter') 		
	
	## write excel handle
	df_results.to_excel(writer, sheet_name='results') 					## write results
	df_identified.to_excel(writer, sheet_name='identified') 			## Identified genes: ARIBA flags explained
	information_ontology.to_excel(writer, sheet_name='CARD_ontology') 	## CARD ontology
	original_data.to_excel(writer, sheet_name='ARIBA_report') 			## Original data from ARIBA
	summary_data.to_excel(writer, sheet_name='ARIBA_summary') 			## ARIBA summary generated
	fileFlags_data.to_excel(writer, sheet_name='flags') 				## ARIBA flags explained
	
	name_csv = folder + '/' + sampleName + '_CARD_summary.csv'
	df_identified.to_csv(name_csv)
	
	## close excel handle
	writer.save()														

	return (name_excel, name_csv)

########################################
def found_results(colnames, dataFrame, list_found_genes, db2use_name):
	##
	## For the information provided by ARIBA summary, get only genes that confer resistance/virulence
	## according to the database due to presence or identified variants
	##
	
	## create dataframe for parsing results and later printing
	df_results = pd.DataFrame(columns=colnames, index=list_found_genes)
	
	## loop
	cluster = dataFrame.groupby(['cluster'])
	for name, group in cluster:
		
		## get ID according to database
		info = get_id(db2use_name, group)
		
		if (db2use_name == 'CARD'):
			df_results.loc[name]['ID'] = info['ID']		
			df_results.loc[name]['Reference'] = info['reference']				

		elif (db2use_name == 'VFDB'):
			df_results.loc[name]['ID'] = info['ID']
			df_results.loc[name]['Reference'] = info['protein_id']
			df_results.loc[name]['Description'] = info['name_id']
			df_results.loc[name]['Species'] = info['species_id']
		else:
			df_results.loc[name]['ID'] = info['ID']		
		
		## coding or non-coding
		sum_gene = int(group['gene'].sum())
		if (sum_gene >= 1):
			df_results.loc[name]['Protein-coding'] = 'yes'
		else:
			df_results.loc[name]['Protein-coding'] = 'no'
			
		## presence_absence
		summary_var_Only=0
		for var_Only in group['var_only']:
			if (var_Only == '.'):
				var_Only = 0
			summary_var_Only = summary_var_Only + int(var_Only)			

		if (summary_var_Only >= 1):
			df_results.loc[name]['Presence/Absence'] = 'no'
		else:
			df_results.loc[name]['Presence/Absence'] = 'yes'

		## sum types: has_known_var
		summary = 0
		for var_name in group['has_known_var']:
			if (var_name == '.'):
				var_name = 0
			summary = summary + int(var_name)		

		if (summary >= 1):
			hash_known_var_data = group.loc[group['has_known_var'] == '1']
			df_results.loc[name]['Variants'] = "; ".join(list(hash_known_var_data['known_var_change']))
			
			var_description = list(hash_known_var_data['var_description'])
			desc_list = []
			for desc in var_description:
				text = desc.split(":.:")[1]
				desc_list.append(text)
			df_results.loc[name]['Additional information'] = "; ".join(set(desc_list))
			
		else:
			df_results.loc[name]['Variants'] = summary
			df_results.loc[name]['Additional information'] = '-'

		## other
		description = "; ".join(set(list(group['free_text'])))
		description = description.replace('b\'', '')
		df_results.loc[name]['Description'] = description

	## debugging
	#pd.set_option('display.max_colwidth', -1)
	#pd.set_option('display.max_columns', None)
	#print (df_results)

	return (df_results)

########################################
def get_id(db2use_name, group):
	## 
	## Retrieve gene ID of a given CARD, VFDB or any other database
	## 
	info = {}

	### Get ID according to Database
	if db2use_name == 'CARD':
		ariba_ref_name = str(group['ref_name']).split('.')
		info['ID'] = 'ARO:' + ariba_ref_name[1]
		info['reference'] = ariba_ref_name[2]

	elif db2use_name == 'VFDB':
		## Get VFDB id
		text_free = "".join(set(list(group['free_text'])))
		text_search = re.search(r"Original name:\s(VF.*\d+)\((.*\d+)\)\s(\(.*)\[(.*)\]", text_free)
		if text_search:
			info['ID'] = text_search.group(1)
			info['protein_id'] = text_search.group(2)
			info['name_id'] = text_search.group(3)
			info['species_id'] = text_search.group(4)
		else:
			info['ID'] = '-'
			info['protein_id'] = '-'
			info['name_id'] = text_free
			info['species_id'] = '-'

	else: 
		## different database
		info['ID'] = '-'
	
	return (info)

########################################
def identified_results(original_data, db2use_name, list_found_genes, assembly_threshold):

	######################################################################################	
	## analyze each cluster identified: might or might not confer resistance/virulence
	######################################################################################
	##
	## For the information provided by ARIBA original data, get all genes identified
	## (might produce or not resistance/virulence due to presence or identified variants)
	##
	## Generate tag:
	##	identified: identified gene as confering resistance/virulence
	##	found: found gene and assembly but not confering resistance/virulence
	##	partial: not fulfilling > assembly_threshold % 		
	## 
	found_genes_cluster = original_data.groupby(['cluster'])
	index_names = set(original_data['cluster'].to_list())

	## create dataframe for parsing results and later printing
	colnames_identified = ['Status', 'ID', 'Protein-coding','Presence_Absence', 'Variants', 'pc_ident', 'pc_len']
	df_identified = pd.DataFrame(columns=colnames_identified, index=index_names)
	df_identified.index.names = ['Genes']

	## loop
	for name, group in found_genes_cluster:

		## get ID according to database
		info = get_id(db2use_name, group)
		df_identified.loc[name]['ID'] = info['ID']		
	
		## coding or non-coding
		sum_gene = int(group['gene'].sum())
		if (sum_gene >= 1):
			df_identified.loc[name]['Protein-coding'] = 'yes'
		else:
			df_identified.loc[name]['Protein-coding'] = 'no'
			
		## presence_absence
		summary_var_Only=0
		for var_Only in group['var_only']:
			if (var_Only == '.'):
				var_Only = 0
			summary_var_Only = summary_var_Only + int(var_Only)
			
		if (summary_var_Only >= 1):
			df_identified.loc[name]['Presence_Absence'] = 'no'
		else:
			df_identified.loc[name]['Presence_Absence'] = 'yes'

		## sum types: has_known_var
		summary = 0
		for var_name in group['has_known_var']:
			if (var_name == '.'):
				var_name = 0
			summary = summary + int(var_name)		

		if (summary >= 1):
			df_identified.loc[name]['Variants'] = 'yes'
		else:
			df_identified.loc[name]['Variants'] = 'no'
			
		## pc_ident
		pc_ident = group['pc_ident'].to_list()
		df_identified.loc[name]['pc_ident'] = pc_ident[0]

		## pc_len = (ref_base_assembled / ref_len)*100
		ref_base = group['ref_base_assembled'].to_list()
		ref_len = group['ref_len'].to_list()
		pc_float = float(ref_base[0]/ref_len[0])
		pc_len = pc_float*100		
		df_identified.loc[name]['pc_len'] = pc_len
		
		### status
		if any(name in s for s in list_found_genes):
			#print ("Gene: ", name, "-- Identify to confer resistance")
			df_identified.loc[name]['Status'] = 'Identified'
		else:
			if pc_float > float(assembly_threshold):
				#print ("Gene: ", name, "-- Found, but not confering resistance.")
				df_identified.loc[name]['Status'] = 'Found'
			else:
				#print ("Gene: ", name, "-- Found, but not confering resistance -- Partial.")
				df_identified.loc[name]['Status'] = 'Partial'
	
	### debugging
	##print (df_identified)
	return (df_identified)
	
#############################################################
def parse_results(folder, sampleName, fileResults, fileFlags, summary):
	## [TODO]
	print (colored("\n\n***** TODO: Implement if a different database provided (!= CARD, VFDB) *****\n\n", 'red'))
	#return(name_excel, name_csv)

#############################################################
def check_results(db2use, outdir_sample, assembly_cutoff, card_trick_info):
	
	"""
	.. seealso:: Additional information to ARIBA results generated.
	
		- :ref:`ARIBA-explained`
	
	"""
	
	## 
	## outdir_sample is a dataframe containing information of the output folder generated by ariba. 
	## It is index for each database and for each sample.
	## This function iterates for each sample and generates call to specific function to parse results.
	## 

	## iterate multi-index dataframe
	dataFrame_results = pd.DataFrame(columns=("csv", "excel", "database"))
	for sample, data in outdir_sample.groupby(level='sample'):
		for database, data2 in data.groupby(level='db'): 
			if (database != db2use):
				continue

			folderResults = data2.loc[sample, db2use]['output']
			outfolder = data2.loc[sample, db2use]['dirname']
			if db2use.endswith('card_prepareref/'):
				database = 'card'
				name_db = 'CARD'
			elif db2use.endswith('vfdb_full_prepareref/'):
				database = 'vfdb_full'
				name_db = 'VFDB'
			else:
				database = 'other'
				name_db = 'other'

			## might generate conflicts if several other databases provided
			## TODO: check
			filename_stamp = outfolder + '/.success_' + database
			if os.path.isfile(filename_stamp):
				stamp =	functions.read_time_stamp(filename_stamp)
				print (colored("\tA previous command generated results on: %s [%s]" %(stamp, sample), 'yellow'))
				name_excel = outfolder + '/' + sample + '_' + name_db + '_results.xlsx'
				name_csv = outfolder + '/' + sample + '_' + name_db + '_summary.csv'

			else:
				(name_excel, name_csv) = results_parser(database, folderResults, sample, outfolder, assembly_cutoff, card_trick_info)

			dataFrame_results.loc[sample] = (name_csv, name_excel, name_db) ## to return

	return (dataFrame_results)

##################################################################
def results_parser(database, folderResults, sampleName, outfolder, assembly_cutoff, card_trick_info):	
	"""Parse ARIBA results
	
	This function basically extracts files and generated additionally information for later
	parse according to type of database provided.
	
	.. seealso:: Additional information to ARIBA results generated.
	
		- :ref:`ARIBA-explained`
	"""	
	if not os.path.exists(folderResults):
		print ("+ Finish parsing information for sample [%s]. Results folder does not exist." %sampleName)
		return('NaN','NaN')
		
	## get files
	list_files = os.listdir(folderResults)

	## init
	assemblies=""
	assemled_genes=""
	fileResults=""

	print ("\n+ Parsing result file for sample: ", sampleName)

	## extract files
	print ("\n+ Extracting files if necessary:")
	for f in list_files:
		filePath = os.path.join(folderResults, f)
		if f.endswith('.gz'):
			functions.extract(filePath, folderResults)
		if (f=='report.tsv'):
			fileResults=filePath
		elif (f=='assemblies.fa.gz'):
			assemblies=os.path.join(folderResults, 'assemblies.fa')
		elif (f=='assembled_genes.fa.gz'):
			assemled_genes=os.path.join(folderResults, 'assembled_genes.fa')
	print ("\n")
	
	## no results generated
	if not functions.is_non_zero_file(fileResults):
		print('+ No results generated for sample: ', sampleName)
		return('','')
		
	### expand flags
	flagResults = folderResults + '/flags_explain.tsv'
	fileFlags = ariba_caller.ariba_expandflag(fileResults, flagResults)
	
	######################
	## generate summary
	######################
	##
	## ariba has function that generates a summary for samples
	##
	summary_results_tmp = folderResults + '/report_summary_tmp'
	summary_results = folderResults + '/report_summary.csv'
	options = "--no_tree"
		## Info
		## https://github.com/sanger-pathogens/ariba/wiki/The-assembled-column-from-ariba-summary
	
	ariba_caller.ariba_summary(summary_results_tmp, [fileResults], options)
	
	## fix names: just for aesthetics
	fake_dict = {sampleName : fileResults}
	ariba_caller.fix_ariba_summary(summary_results_tmp + '.csv', summary_results, fake_dict)
	os.remove(summary_results_tmp + '.csv')
	
	############################################
	### check results according to database
	############################################
	if (database == 'vfdb_full'):
		(name_excel, name_csv) = parse_vfdb(outfolder, sampleName, fileResults, fileFlags, summary_results, assembly_cutoff)
	elif (database == 'card'):
		(name_excel, name_csv) = parse_card(outfolder, sampleName, fileResults, fileFlags, summary_results, assembly_cutoff, card_trick_info)
	else: 
		## [TODO] check results according to databases different than CARD/VFDB
		(name_excel, name_csv) = parse_results(outfolder, sampleName, fileResults, fileFlags, summary_results)				
	
	print ('\tCheck additional information on ', name_excel)
	
	## print success timestamp
	filename_stamp = outfolder + '/.success_' + database
	stamp =	functions.print_time_stamp(filename_stamp)

	return (name_excel, name_csv)

#############################################################
def check_VFDB(VFDB_path):
	## 
	## Given a folder, check/download VFDB information
	## and print information.
	## Returns files downloaded or available
	## 
	
	## download
	download_VFDB_files(VFDB_path)

	## Print info files
	files = os.listdir(VFDB_path)
	VFDB_files = {}
	print ('+ Several files are available with multiple information:')
	for item in files:
		if (item == 'VFs.xls'):
			VFDB_files['VFs-xls'] = VFDB_path + '/' + item
			print ('\tVFs.xls: Virulence Factor description file')
		if (item == 'Comparative_tables_from_VFDB'):
			VFDB_files['Comparative_tables'] = VFDB_path + '/' + item
			print ('\tComparative_tables_from_VFDB: Intra-genera comparison tables')
	print ('\nFor further details please visit: <http://www.mgc.ac.cn/VFs/main.htm>')
	return (VFDB_files)
	
#############################################################
def download_VFDB_files(folder):
	## 
	## Given a folder, check if it contains VFDB information
	## or download it from website: http://www.mgc.ac.cn
	## 
	links = ("http://www.mgc.ac.cn/VFs/Down/VFs.xls.gz","http://www.mgc.ac.cn/VFs/Down/Comparative_tables_from_VFDB.tar.gz")	

	## check if data is downloaded, how old is the data and if it is necessary to download again
	## consider >30 days long enough to be updated again
	
	## time stamp
	filename_stamp = folder + '/download_timestamp.txt' 
	if os.path.exists(folder):
		if os.path.isfile(filename_stamp):
			stamp = functions.read_time_stamp(filename_stamp)
			print ("+ A previous download generated results on: ", stamp)
			days_passed = functions.get_diff_time(stamp)
			print ("+ %s days ago" %days_passed)		
			if (days_passed > 30): ## download again
				print ("+ Downloading information again just to be sure...")
			else:
				print ("+ No need to download data again.")
				return()
	else:
		functions.create_folder(folder)
	
	## Open file and readlines
	print ('+ Downloading files:\n')
	for line in links:
		if not line.startswith('#'):  
			functions.wget_download(line, folder)
	
	## decompress files			
	print ('+ Decompressing gzip files\n')
	files = os.listdir(folder)
	for item in files:
		#print (folder)
		if item.endswith('.gz'):
			functions.extract(folder + '/' + item, folder)

	## make stamp time
	functions.print_time_stamp(filename_stamp)
	
	return()

#############################################################
def subsetting_species(species, folder, excel_file):
	## 
	## Given a Species name, this function subsets information from VFDB
	## and generates excel file from input name provided (excel_file)
	## 	

	print ('+ Subsetting information for species of interest')
	subset_xlsx_file = folder + '/VFs_' + species + '.xls'
	
	## subset
	dfs = pd.read_excel(excel_file, sheet_name="VFs", index_cols=0, skiprows=1)
	df2 = dfs.set_index("Bacteria", drop = False) 				## order
	df_species = df2.loc[species:species] 						## filter by species
	df_species2 = df_species.set_index("VFID", drop = False) 	## order
	ls_items = list(df_species2)								## get column headers
	df_species2.to_excel(subset_xlsx_file, sheet_name=species, header=ls_items)

######
def main():
	## this code runs when call as a single script

  	## control if options provided or help
	if len(sys.argv) > 1:
		print ("")
	else:
		help_options()
		exit()
		
	#species = argv[1]
	#folder = os.path.abspath(argv[2])
	#VFDB_path = os.path.abspath(argv[3])
	
	## Start
	## check if files are downloaded first	
	#VFDB_files = check_VFDB(VFDB_path)
	
	## subset for genera of interest
	#functions.create_folder(folder)
	
	# subset for species of interest	
	#subsetting_species(species, folder, VFDB_files['VFs-xls'])

	#sampleName = "test_sample"
	#folder = argv[4]
	#parse_vfdb(folder, sampleName, argv[1], argv[2], argv[3])
	
######
def help_options():
	print ("\nUSAGE: python %s species path VFDB_path\n"  %os.path.realpath(__file__))
		
######
'''******************************************'''
if __name__== "__main__":
	main()

## http://www.mgc.ac.cn/cgi-bin/VFs/gene.cgi?GeneID=VFG049753
## http://www.mgc.ac.cn/cgi-bin/VFs/vfs.cgi?VFID=VF0403#VF0403



import os
import shutil
import sys
import glob
from setuptools import setup, find_packages, Extension

#######
def get_require_modules():
    """
    Get main python requirements modules
    """
    with open("./BacterialTyper/config/python/python_requirements_summary.csv", 'r') as f:
        myModules = [line.strip().split(',')[0] for line in f]
    
    return myModules

#######
def get_version():
    """
    Original code: PhiSpy setup.py 
    https://github.com/linsalrob/PhiSpy/blob/master/setup.py
    """
    with open("./VERSION", 'r') as f:
        v = f.readline().strip()
    return v

### subdmodules:: Islandpath-DIMOB
#islandpath_files = [os.path.join('third_party', 'Islandpath-DIMOB', x) for x in islandpath_files]
#islandpath_mod = Extension(
#    "Islandpath-DIMOB",
#    islandpath_files,
#    include_dirs=[os.path.join('third_party', 'Islandpath-DIMOB')],
#)
#######
long_description_text = ""
with open("README.md", "r") as fh:
    long_description_text = fh.read()

#######
setup(
    name='BacterialTyper',
    version=get_version(),

    scripts=glob.glob('main/*'),
    author="Jose F. Sanchez-Herrero",
    author_email="jfbioinformatics@gmail.com",

    long_description_content_type="text/markdown",
    long_description=long_description_text,

    url="https://github.com/HCGB-IGTP/BacterialTyper/",

    install_requires=get_require_modules(),
    packages = find_packages(),

    license='MIT License',    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)

## https://github.com/sanger-pathogens/ariba/blob/master/setup.py
## https://dzone.com/articles/executable-package-pip-install
## https://github.com/BillMills/pythonPackageLesson



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Trimms sequence adapters within fastq reads.
"""
## import useful modules
import os
import sys
import re
import time
from io import open
import shutil
import concurrent.futures
from termcolor import colored

## import my modules
from BacterialTyper.scripts import trimmomatic_call
from BacterialTyper.scripts import multiQC_report
from BacterialTyper.scripts import sampleParser
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.modules import help_info
from BacterialTyper.modules import qc
from BacterialTyper.data import data_files

##############################################
def run(options):

	## init time
	start_time_total = time.time()

	##################################
	### show help messages if desired	
	##################################
	if (options.help_format):
		## help_format option
		sampleParser.help_format()
		exit()
	elif (options.help_trimm_adapters):
		## help on trimm adapters
		trimmomatic_call.print_help_adapters()
		exit()
	elif (options.help_project):
		## information for project
		help_info.project_help()
		exit()
	elif (options.help_multiqc):
		## information for Multiqc
		multiQC_report.multiqc_help()
		exit()
		
	## debugging messages
	global Debug
	if (options.debug):
		Debug = True
	else:
		Debug = False
		
	### set as default paired_end mode
	if (options.single_end):
		options.pair = False
	else:
		options.pair = True
	
	functions.pipeline_header()
	functions.boxymcboxface("Trimming samples")
	print ("--------- Starting Process ---------")
	functions.print_time()

	## absolute path for in & out
	input_dir = os.path.abspath(options.input)
	outdir=""

	## Project mode as default
	if (options.detached):
		options.project = False
		outdir = os.path.abspath(options.output_folder)
	else:
		options.project = True
		outdir = input_dir	
	
	## get files
	pd_samples_retrieved = sampleParser.get_files(options, input_dir, "fastq", ("fastq", "fq", "fastq.gz", "fq.gz"))
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
		print (pd_samples_retrieved)

	## generate output folder, if necessary
	print ("\n+ Create output folder(s):")
	if not options.project:
		functions.create_folder(outdir)
	## for samples
	outdir_dict = functions.outdir_project(outdir, options.project, pd_samples_retrieved, "trimm")
	
	## optimize threads
	name_list = set(pd_samples_retrieved["name"].tolist())
	threads_job = functions.optimize_threads(options.threads, len(name_list)) ## threads optimization
	max_workers_int = int(options.threads/threads_job)

	## debug message
	if (Debug):
		print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
		print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
		print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))

	print ("+ Trimming adapters for each sample retrieved...")	
	
	# Group dataframe by sample name
	sample_frame = pd_samples_retrieved.groupby(["name"])
	
	# Trimming adapters
	if (options.adapters):
		# Adapter file provided
		options.adapters = os.path.abspath(options.adapters)
		print("\t- Adapters file provided...")
	else:
		# Get default adpaters file
		print("\t- Default Trimmomatic adapters (v0.39) will be used...")
		options.adapters = data_files.data_list("available_Trimmomatic_adapters")
		
	## send for each sample
	with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor:
		commandsSent = { executor.submit(trimmo_caller, sorted(cluster["sample"].tolist()), outdir_dict[name], name, threads_job, Debug, options.adapters): name for name, cluster in sample_frame }

		for cmd2 in concurrent.futures.as_completed(commandsSent):
			details = commandsSent[cmd2]
			try:
				data = cmd2.result()
			except Exception as exc:
				print ('***ERROR:')
				print (cmd2)
				print('%r generated an exception: %s' % (details, exc))

	print ("\n\n+ Trimming samples has finished...")
	## functions.timestamp
	start_time_partial = functions.timestamp(start_time_total)

	## get files generated and generate symbolic link
	if not options.project:
		dir_symlinks = functions.create_subfolder('link_files', outdir)
		files2symbolic = []
		folders = os.listdir(outdir)

		## debug message
		if (Debug):
			print (colored("**DEBUG: generate symbolic links for each file in " + dir_symlinks + "**", 'yellow'))
		
		for fold in folders:
			if fold.endswith(".log"):
				continue
			else:
				this_folder = outdir + '/' + fold
				subfiles = os.listdir(this_folder)
				for files in subfiles:
					files_search = re.search(r".*trim_R\d{1}.*", files) ## only paired-end. Todo: single end
					if files_search:
						files2symbolic.append(this_folder + '/' + files)
	
		functions.get_symbolic_link(files2symbolic, dir_symlinks)

	if (options.skip_report):
		print ("+ No report generation...")
	else:
		print ("\n+ Generating a report using MultiQC module.")
		outdir_report = functions.create_subfolder("report", outdir)
	
		## call multiQC report module
		givenList = [ v for v in outdir_dict.values() ]
		my_outdir_list = set(givenList)
		
		## debug message
		if (Debug):
			print (colored("\n**DEBUG: my_outdir_list for multiqc report **", 'yellow'))
			print (my_outdir_list)
			print ("\n")

		trimm_report = functions.create_subfolder("trimm", outdir_report)
		multiQC_report.multiQC_module_call(my_outdir_list, "Trimmomatic", trimm_report,"")
		print ('\n+ A summary HTML report of each sample is generated in folder: %s' %trimm_report)
		
		## create fastqc for trimmed reads
		pd_samples_retrieved_trimmed = sampleParser.get_files(options, input_dir, "trim", ['_trim'])
		qc.fastqc(pd_samples_retrieved_trimmed, outdir, options, start_time_partial, "trimmed", Debug)
		
	print ("\n*************** Finish *******************")
	start_time_partial = functions.timestamp(start_time_total)
	print ("\n+ Exiting trimm module.")
	return()
	

#############################################
def trimmo_caller(list_reads, sample_folder, name, threads, Debug, adapters):
	## check if previously assembled and succeeded
	filename_stamp = sample_folder + '/.success'
	if os.path.isfile(filename_stamp):
		stamp =	functions.read_time_stamp(filename_stamp)
		print (colored("\tA previous command generated results on: %s [%s]" %(stamp, name), 'yellow'))
	else:
		# Call trimmomatic
		trimmomatic_call.trimmo_module(list_reads, sample_folder, name, threads, Debug, adapters)



	



__all__ = [
    'annotation',
    'ariba_caller',
    'bacteriophage',
    'blast_parser',
    'BUSCO_caller',
    'card_trick_caller',
    'database_generator',
    'database_user',
    'edirect_caller',
    'fastqc_caller',
    'functions',
    'genomic_island',
    'MLSTar',
    'min_hash_caller',
    'multiQC_report',
    'plasmidID_call',
    'phylo_parser',
    'sampleParser',
    'spades_assembler',
    'species_identification_KMA',
    'trimmomatic_call',
    'variant_calling',
    'virulence_resistance'
]

from BacterialTyper.scripts import *


#!/usr/bin/env python3
"""
Calls IslandPath Dimob software for identification of putative genomics islands within a genome.
"""
## useful imports
from sys import argv
import time
import io
import os
import re
import sys
import pandas as pd
import shutil
from termcolor import colored

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config


######
def GI_module(genbank_file, name, outdir, Debug, cutoff_dinuc_bias=8, min_length=1000):
    """Identify genomic islands (GI) within the genbank file provided. They are calculated
    based on gene annotation and dinucleotide bias region using the software `IslandPath-DIMOB`_.
    
    :param genbank_file: Absolute path to annotation file in Genbank format.
    :param name: Sample identifier. 
    :param outdir: Absolute path to output folder.
    :param cutoff_dinuc_bias: Dinucleotide bias cutoff
    :param min_length: Minimun length for the regions to be reported

    :type name: string
    :type genbank_file: string
    :type outdir: string
    :type cutoff_dinuc_bias: int
    :type min_length: int

    The Dimob.pl perl script has two mandatory argument which are the input :file:`genbank_file` and an output name.
    
    .. code-block:: sh

        Usage:
        perl Dimob.pl <genome.gbk> <output_name> [cutoff_dinuc_bias] [min_length]
        
        Default values:
            cutoff_dinuc_bias = 8
            min_length = 8000
        
        Example:
            perl Dimob.pl example/NC_003210.gbk NC_003210_GIs
            perl Dimob.pl example/NC_003210.gbk NC_003210_GIs 6 10000
            perl Dimob.pl example/NC_000913.embl NC_000913_GIs 6 10000

    
    During the development of BacterialTyper, we generated a modification of the original `IslandPath-DIMOB`_ to analyze 
    contig sequence data and generated different output format for better clarificaiton and interpretaion of results. 
    We forked the original code into a new git repository and update the code accordingly. See details here: https://github.com/JFsanchezherrero/islandpath.
    
     .. include:: ../../links.inc
    
    """
    
    ## filename stamp of the process
    filename_stamp = outdir + '/.Dimob'

    # check if previously done
    if os.path.isfile(filename_stamp):
        stamp = functions.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- Dimob]" %(stamp, name), 'yellow'))
    else:    
        ## debug message
        if (Debug):
            print (colored("**DEBUG: Call Dimob for sample %s " %name + "**", 'yellow'))
            print ("genbank_file", genbank_file)
            print ("outdir: ", outdir)
           
        ## Call IslandPath Dimob executable perl file.
        dimob_pl = set_config.get_exe("dimob", Debug)
        perl_exe = set_config.get_exe("perl", Debug)
        
        ## command
        outdir_sample = os.path.join(outdir, name)
        log_file = outdir_sample + '.log'
        perl_cmd = '%s %s %s %s %s %s > %s' %(perl_exe, dimob_pl, genbank_file, outdir_sample, cutoff_dinuc_bias, min_length, log_file)
    
        code = functions.system_call(perl_cmd)
        ##
        if code:
            ## when finished print time stamp in  output + '/.Dimob'
            stamp = functions.print_time_stamp(filename_stamp)
        else:
            return False
    
    return (outdir)

def help_Dimob():
    print (colored("\n\n***** TODO: Generate this help message *****\n\n", 'red'))
    
def help_options():
    print (colored("\n\n***** TODO: Generate this help message *****\n\n", 'red'))
    
def main():
    ## control if options provided or help
    if len(sys.argv) != 4:
        help_options()
        exit()
    
    ## argv
    gbk_file = os.path.abspath(argv[1])
    outdir = os.path.abspath(argv[2])
    name=argv[3]
    
    ## call
    print ("\n+ Generate annotation of putative Genomic Islands for sample %s in folder: %s" %(name, outdir))
    GI_module(gbk_file, name, outdir, True)    
    
######
'''******************************************'''
if __name__== "__main__":
    main()
    



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Calls Entrez Direct NCBI search.
'''
## useful imports
import time
import io
import os
import re
import sys
import pandas as pd
from termcolor import colored

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config

## additional information:
## https://ncbi-hackathons.github.io/EDirectCookbook/

####################################
def help_edirect():
	"""
	Prints edirect_caller script help options
	"""
	functions.boxymcboxface("ENTREZ DIRECT (EDirect)")
	print ("+ Connects with NCBI entrez databases")
	print ("+ Searches, retrieves, and parses data from NCBI databases through the Unix command line")
	print ("+ Provides access to the NCBI's suite of interconnected databases")
	print ("+ Includes an argument-driven function that simplifies the extraction of \ndata from document summaries or other results that are in structured XML format.")
	print ("\n")

###############
def generate_docsum_call(db, query, outfile):
	esearch_bin = set_config.get_exe("esearch") 
	efetch_bin = set_config.get_exe("efetch")
	return(docsum_call(db, query, outfile, esearch_bin, efetch_bin))

###############
def docsum_call(db, query, outfile, esearch_bin, efetch_bin):
	cmd = ("%s -db %s -query %s | %s -format docsum > %s" %(esearch_bin, db, query, efetch_bin, outfile))
	return(functions.system_call(cmd))
	
###############
def generate_xtract_call(docsum_file, pattern, element, outfile):
	xtract_bin = set_config.get_exe("xtract") 
	return(xtract_call(docsum_file, pattern, element, outfile, xtract_bin))
	
###############
def xtract_call(docsum_file, pattern, element, outfile, xtract_bin):
	cmd = ("cat %s | %s -pattern %s -sep ',' -element %s > %s" %(docsum_file, xtract_bin, pattern, element, outfile))
	return(functions.system_call(cmd))	

###############
def generate_seq_search_call(db, query, outfile, revcomp, start=0, end=-1, format='fasta'):

	## Sequence Range
	##   -seq_start     First sequence position to retrieve
	##   -seq_stop      Last sequence position to retrieve
	##   -strand        1 = forward DNA strand, 2 = reverse complement
	##   -revcomp       Shortcut for strand 2

	efetch_bin = set_config.get_exe("efetch")
	cmd = ("%s -db %s -id %s -seq_start %s -seq_stop %s -format %s" %(efetch_bin, db, query, start, end, format))
	
	## add reverse complement
	if (revcomp):
		cmd = cmd + ' -revcomp'

	## add output file
	cmd = cmd + ' > %s' %outfile
		
	return(functions.system_call(cmd))

## docsum Nucleotide entry
##esearch -db nuccore -query NZ_CP029083.1 | efetch -format docsum > docsum_nucleotide_entry.txt
##cat docsum_nucleotide_entry.txt | xtract -pattern DocumentSummary -element BioSample > BioSample.txt

## Docsum BioSample
##esearch -query BioSample.txt -db assembly | efetch -format docsum > BioSample_docsum.txt
##cat BioSample_docsum.txt | xtract -pattern DocumentSummary -element FtpPath_GenBank




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Initiates, updates and configures database.
"""
## useful modules
import time
import os
import sys
from termcolor import colored
import pandas as pd

## import my modules
from BacterialTyper.scripts import database_generator
from BacterialTyper.scripts import database_user
from BacterialTyper.scripts import ariba_caller
from BacterialTyper.scripts import multiQC_report
from BacterialTyper.scripts import species_identification_KMA
from BacterialTyper.config import set_config 
from BacterialTyper.scripts import functions
from BacterialTyper.scripts import BUSCO_caller

###############################################################
def run_database(options):

	## init time
	start_time_total = time.time()
	start_time_partial = start_time_total

	## debugging messages
	global Debug
	if (options.debug):
		Debug = True
		print ("[Debug mode: ON]")
	else:
		Debug = False

	## message header
	functions.pipeline_header()
	functions.boxymcboxface("Database")
	print ("--------- Starting Process ---------")
	functions.print_time()

	kma_bin = set_config.get_exe("kma")

	######################################################
	## print further information if requested
	if (options.help_ARIBA):
		print ("ARIBA databases information:")	
		ariba_caller.help_ARIBA()
		exit()

	elif (options.help_BUSCO):
		BUSCO_caller.print_help_BUSCO()
		exit()
		
	elif (options.help_KMA):
		species_identification_KMA.help_kma_database()
		exit()
	######################################################

	## create folder
	## absolute
	options.path = os.path.abspath(options.path)	
	functions.create_folder(options.path)

	#########
	if Debug:
		print (colored("DEBUG: absolute path folder: " + options.path, 'yellow'))

	##########
	## NCBI	##
	##########
	## if any NCBI options provided
	if any ([options.ID_file, options.descendant]):
		## create folders
		NCBI_folder = functions.create_subfolder('NCBI', options.path)
		if (options.ID_file):
			## get path and check if it is file
			abs_path_file = os.path.abspath(options.ID_file)
			if os.path.isfile(abs_path_file):
				print ()
				functions.print_sepLine("*",50, False)
				print ("--------- Check NCBI ids provided ---------\n")
				functions.print_sepLine("*",70, False)
				## get file information
				print ("\t+ Obtaining information from file: %s" %abs_path_file)
				strains2get = functions.get_data(abs_path_file, ',', '')
				dataBase_NCBI = database_generator.NCBI_DB(strains2get, NCBI_folder, Debug)

				#########
				if Debug:
					print (colored("DEBUG: NCBI data provided: ", 'yellow'))
					print (options.ID_file)

				## functions.timestamp
				start_time_partial = functions.timestamp(start_time_partial)
				## strains downloaded would be included to a kma index

		## Get all entries belonging to this taxon provided
		if (options.descendant):
			#########
			if Debug:
				print (colored("DEBUG: NCBI descendant option: ON ", 'yellow'))
			
			print ()
			functions.print_sepLine("*",70, False)
			print ("--------- Check descendant NCBI taxonomy ids provided ---------\n")
			functions.print_sepLine("*",70, False)
			## [TODO]
			dataBase_NCBI = database_generator.NCBI_descendant(options.descendant, NCBI_folder, Debug)
			
		##############################################################
		## update KMA database with NCBI information retrieved
		##############################################################
		print ('\n\n+ Update database for later identification analysis...')
		list_of_files = dataBase_NCBI['genome'].tolist()
		kma_db = functions.create_subfolder('KMA_db', options.path)	
		genbank_kma_db = functions.create_subfolder('genbank', kma_db)	
		
		print ('+ Database to update: ', genbank_kma_db)
		species_identification_KMA.generate_db(list_of_files, 'genbank_KMA', genbank_kma_db, 'new', 'batch', Debug, kma_bin)

		## time stamp
		start_time_partial = functions.timestamp(start_time_total)

	###############
	## user_data ##
	###############	
	if options.project_folder:
		
		##
		dataBase_user = pd.DataFrame()	
		## get absolute path
		abs_project_folder = os.path.abspath(options.project_folder)
		if os.path.exists(abs_project_folder):
			#########
			if Debug:
				print (colored("DEBUG: User provides folder containing project", 'yellow'))

			print ()
			functions.print_sepLine("*",70, False)
			print ("--------- Check user provided project folder ---------")
			functions.print_sepLine("*",70, False)
			dataBase_user = database_user.update_database_user_data(options.path, abs_project_folder, Debug, options)
		else:
			print (colored("ERROR: Folder provided does not exists: %s" %options.project_folder, 'red'))
			exit()

		##############################################################
		## update KMA database with user_data information retrieved
		##############################################################
		print ('\n\n+ Update database for later identification analysis...')
		list_of_files = dataBase_user['genome'].tolist()
		kma_db = functions.create_subfolder('KMA_db', options.path)	
		user_kma_db = functions.create_subfolder('user_data', kma_db)	
		
		print ('+ Database to update: ', user_kma_db)
		species_identification_KMA.generate_db(list_of_files, 'userData_KMA', user_kma_db, 'new', 'batch', Debug, kma_bin)
		
		## time stamp
		start_time_partial = functions.timestamp(start_time_total)

	##########
	## ARIBA
	##########
	print ()
	functions.print_sepLine("*",50, False)
	print ("--------- Check ARIBA parameters provided --------")
	functions.print_sepLine("*",50, False)
	if (options.no_ARIBA):
		print ("+ No ARIBA databases would be downloaded...")
		
		#########
		if Debug:
			print (colored("DEBUG: No option ARIBA", 'yellow'))
	
	else:
		#functions.print_sepLine("*",50, False)
		
		### ariba list databases
		ariba_dbs_list = ['CARD', 'VFDB']
		
		if (options.no_def_ARIBA):
			ariba_dbs_list = options.ariba_dbs
		else:
			if (options.ariba_dbs):
				ariba_dbs_list = ariba_dbs_list + options.ariba_dbs
				ariba_dbs_list = set(ariba_dbs_list)
			
		#########
		if Debug:
			print (colored("DEBUG: Option ARIBA", 'yellow'))
			print (options.ariba_dbs)
		
		ariba_caller.download_ariba_databases(ariba_dbs_list, options.path, Debug, options.threads)
	
		### ariba list databases		
		if (options.ariba_users_fasta):
			print ("+ Generate ARIBA database for databases provided: prepare fasta and metadata information")

			#########
			if Debug:
				print (colored("DEBUG: Option user ARIBA db", 'yellow'))
				print (ariba_users_fasta)
				print (ariba_users_meta)

			## [TODO]:	
			## ariba prepareref fasta and metadata

		### timestamp
		start_time_partial = functions.timestamp(start_time_partial)					

	#########
	## kma ##
	#########
	print ()
	functions.print_sepLine("*",50, False)
	print ("--------- Check KMA parameters provided ----------")
	kma_database = options.path + '/KMA_db'	
	functions.create_folder(kma_database)
	
	## types: bacteria, archaea, protozoa, fungi, plasmids, typestrains
	## downloads all "bacterial" genomes from KMA website
	## kma: ftp://ftp.cbs.dtu.dk/public/CGE/databases/KmerFinder/version/

	print ("+ Retrieving information from: ftp://ftp.cbs.dtu.dk/public/CGE/databases/KmerFinder website")		

	## KMA databases to use	
	## only user dbs	
	if (options.no_def_kma):
		if (options.kma_dbs):
			print("+ Only user databases selected will be indexed...")
		else:
			print ("+ No databases selected.")
			print (colored("ERROR: Please select a kma database.", 'red'))
			exit()
			
	## default dbs + user
	else:
		kma_dbs = ["bacteria", "plasmids"]

		## default dbs + user
		if (options.kma_dbs):
			options.kma_dbs = options.kma_dbs + kma_dbs
			options.kma_dbs = set(options.kma_dbs)		
		else:
			options.kma_dbs = kma_dbs

	#########
	if Debug:
		print (colored("DEBUG: options.kma_dbs", 'yellow'))
		print (options.kma_dbs)

	## Get databases
	for db in options.kma_dbs:
		print (colored("\n+ " + db, 'yellow'))
		db_folder = functions.create_subfolder(db, kma_database)		
		species_identification_KMA.download_kma_database(db_folder, db, Debug)

	### timestamp
	start_time_partial = functions.timestamp(start_time_partial)					

	###########	
	## BUSCO ##
	###########
	if (options.BUSCO_dbs):
		print ()
		functions.print_sepLine("*",50, False)
		print ("--------- Check BUSCO datasets provided ---------")
		BUSCO_folder = functions.create_subfolder("BUSCO", options.path)

		#########
		if Debug:
			print (colored("DEBUG: options.BUSCO_dbs", 'yellow'))
			print (options.BUSCO_dbs)

		BUSCO_caller.BUSCO_retrieve_sets(options.BUSCO_dbs, BUSCO_folder)

		### timestamp
		start_time_partial = functions.timestamp(start_time_partial)					

	print ("\n*************** Finish *******************\n")
	start_time_partial = functions.timestamp(start_time_total)

	print ("+ Exiting Database module.\n")
	return()



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Generates sample identification using KMA software and MLSTar. 
Looks for similar entries on GenBank and retrieves them.  
"""
## useful imports
import time
import io
import os
import re
import sys
import concurrent.futures
from termcolor import colored
import pandas as pd

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts import sampleParser
from BacterialTyper.scripts import species_identification_KMA
from BacterialTyper.scripts import database_generator
from BacterialTyper.scripts import MLSTar
from BacterialTyper.scripts import edirect_caller
from BacterialTyper.modules import help_info

####################################
def run_ident(options):
	"""
	Main function acting as an entry point to the module *ident*.
	
	Arguments:
	
	
	.. seealso:: Additional information to PubMLST available datasets.
	
		- :doc:`PubMLST datasets<../../../data/PubMLST_datasets>`
	
	
	"""

	##################################
	### show help messages if desired	
	##################################
	if (options.help_format):
		## help_format option
		sampleParser.help_format()
		exit()

	elif (options.help_project):
		## information for project
		help_info.project_help()
		exit()
	
	elif (options.help_KMA):
		## information for KMA Software
		species_identification_KMA.help_kma_database()
		exit()

	elif (options.help_MLSTar):
		## information for KMA Software
		MLSTar.help_MLSTar()
		exit()
	
	## init time
	start_time_total = time.time()

	## debugging messages
	global Debug
	if (options.debug):
		Debug = True
	else:
		Debug = False
		
	### set as default paired_end mode
	if (options.single_end):
		options.pair = False
	else:
		options.pair = True

	### species_identification_KMA -> most similar taxa
	functions.pipeline_header()
	functions.boxymcboxface("Species identification")

	print ("--------- Starting Process ---------")
	functions.print_time()

	## absolute path for in & out
	input_dir = os.path.abspath(options.input)
	outdir=""

	## Project mode as default
	global Project 
	
	if (options.detached):
		options.project = False
		project_mode=False
		outdir = os.path.abspath(options.output_folder)
		Project=False
	else:
		options.project = True
		outdir = input_dir
		Project=True

	## get files
	pd_samples_retrieved = sampleParser.get_files(options, input_dir, "trim", ['_trim'])
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
		print (pd_samples_retrieved)
	
	## generate output folder, if necessary
	print ("\n+ Create output folder(s):")
	if not options.project:
		functions.create_folder(outdir)
	## for each sample
	outdir_dict = functions.outdir_project(outdir, options.project, pd_samples_retrieved, "ident")	
	
	## let's start the process
	print ("+ Generate an species typification for each sample retrieved using:")
	print ("(1) Kmer alignment (KMA) software.")	
	print ("(2) Pre-defined databases by KMA or user-defined databases.")	
		
	## get databases to check
	retrieve_databases = get_options_db(options)
	
	## time stamp
	start_time_partial = functions.timestamp(start_time_total)
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: retrieve_database **", 'yellow'))
		pd.set_option('display.max_colwidth', None)
		pd.set_option('display.max_columns', None)
		print (retrieve_databases)
	
	######## KMA identification
	dataFrame_kma = KMA_ident(options, pd_samples_retrieved, outdir_dict, retrieve_databases, start_time_partial)
	
	## functions.timestamp
	start_time_partial = functions.timestamp(start_time_partial)
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: retrieve results to summarize **", 'yellow'))
		pd.set_option('display.max_colwidth', None)
		pd.set_option('display.max_columns', None)
		print ("dataframe_kma")
		print (dataFrame_kma)
	
	
	## exit if viral search
	skip=False
	if (len(options.kma_dbs) == 1):
		for i in options.kma_dbs:
			if (i == 'viral'):
				print ()
				MLST_results = ''
				options.fast = True
				skip=True
			
			## what if only plasmids?
			
	## do edirect and MLST if bacteria
	if (not skip):		
		######## EDirect identification
		dataFrame_edirect = edirect_ident(dataFrame_kma, outdir_dict, Debug)
		
		## functions.timestamp
		start_time_partial = functions.timestamp(start_time_partial)
	
		## debug message
		if (Debug):
			print (colored("**DEBUG: retrieve results from NCBI **", 'yellow'))
			pd.set_option('display.max_colwidth', None)
			pd.set_option('display.max_columns', None)
			print ("dataFrame_edirect")
			print (dataFrame_edirect)
			
		######## MLST identification
		MLST_results = MLST_ident(options, dataFrame_kma, outdir_dict, dataFrame_edirect, retrieve_databases)
	
		## functions.timestamp
		start_time_partial = functions.timestamp(start_time_partial)

		## debug message
		if (Debug):
			print (colored("**DEBUG: retrieve results to summarize **", 'yellow'))
			pd.set_option('display.max_colwidth', None)
			pd.set_option('display.max_columns', None)
			print ("MLST_results")
			print (MLST_results)

	## generate summary for sample: all databases
	## MLST, plasmids, genome, etc
	functions.boxymcboxface("Results Summary")
	
	#####################################
	## Summary identification results  ##
	#####################################

	## parse results
	if options.project:
		final_dir = outdir + '/report/ident'
		functions.create_folder(final_dir) 
	else:
		final_dir = outdir

	###
	excel_folder = functions.create_subfolder("samples", final_dir)
	print ('+ Print summary results in folder: ', final_dir)
	print ('+ Print sample results in folder: ', excel_folder)
	
	# Group dataframe results summary by sample name
	sample_results_summary = dataFrame_kma.groupby(["Sample"])

	## debug message
	if (Debug):
		print (colored("**DEBUG: sample_results_summary **", 'yellow'))
		print (sample_results_summary)
	
	##
	results_summary_KMA = pd.DataFrame()
	MLST_all = pd.DataFrame()
	for name, grouped in sample_results_summary:
		
		## create a excel and txt for sample
		name_sample_excel = excel_folder + '/' + name + '_ident.xlsx'
		name_sample_csv = outdir_dict[name] + '/ident_summary.csv' ## check in detached mode

		writer_sample = pd.ExcelWriter(name_sample_excel, engine='xlsxwriter') ## open excel handle
		
		## subset dataframe	& print result
		results_summary_toPrint_sample = grouped[['Sample','#Template','Query_Coverage','Template_Coverage','Depth', 'Database']] 
		results_summary_toPrint_sample.to_excel(writer_sample, sheet_name="KMA") ## write excel handle
		results_summary_toPrint_sample.to_csv(name_sample_csv) ## write csv for sample
		
		## read MLST
		if MLST_results:
			if name in MLST_results:
				sample_MLST = pd.read_csv(MLST_results[name], header=0, sep=',')
				sample_MLST['genus'] = dataFrame_edirect.loc[dataFrame_edirect['sample'] == name, 'genus'].values[0]
				sample_MLST['species'] = dataFrame_edirect.loc[dataFrame_edirect['sample'] == name, 'species'].values[0]
				sample_MLST.to_excel(writer_sample, sheet_name="MLST") ## write excel handle
			
				## Return information to excel
				MLST_all = pd.concat([MLST_all, sample_MLST]) 
			
		## close excel handle
		writer_sample.save() 		

	##
	name_excel = final_dir + '/identification_summary.xlsx'
	print ('+ Summary information in excel file: ', name_excel)
	writer = pd.ExcelWriter(name_excel, engine='xlsxwriter') ## open excel handle

	## KMA dataframe: print result for sources
	results_summary_KMA = dataFrame_kma[['Sample','#Template','Query_Coverage','Template_Coverage','Depth', 'Database']] 
	
	## Sum plasmid and chromosome statistics ##
	## sum coverage
	total_coverage = results_summary_KMA.groupby('Sample')['Query_Coverage'].sum().reset_index()
	
	## debug message
	if (Debug):
		print ("*** Sum: Query_coverage ***")		
		print (total_coverage)
	
	## TODO: FIX SUMMARY REPORT
	results_summary_KMA = results_summary_KMA.set_index('Sample')
	results_summary_KMA = results_summary_KMA.sort_values(by=['Sample', 'Database', 'Query_Coverage'],ascending=[True, True,True])
	results_summary_KMA.to_excel(writer, sheet_name='KMA') ## write excel handle
	
	## write MLST
	if (MLST_results):
		MLST_all.to_excel(writer, sheet_name='MLST')
	
	## write excel and close
	writer.save() ## close excel handle

	print ("\n+ Check summary of results in file generated" )
	
	### timestamp
	start_time_partial = functions.timestamp(start_time_partial)
	
	######################################
	## update database for later usage
	######################################
	if not options.fast:

		functions.boxymcboxface("Update Sample Database")

		## update db
		print ("+ Update database with samples identified")

		## debug message
		if (Debug):
			print (colored("**DEBUG: dataFrame_edirect **", 'yellow'))
			pd.set_option('display.max_colwidth', None)
			pd.set_option('display.max_columns', None)
			print (dataFrame_edirect)

		## dataFrame_edirect
		file_toprint = final_dir + '/edirect_info2download.csv'
		dataFrame_edirect.to_csv(file_toprint)

		## update database with samples identified
		data2download = dataFrame_edirect.filter(['genus','species', 'strain', 'genome'])
		data2download = data2download.rename(columns={'genome': 'NCBI_assembly_ID', 'strain' : 'name'})
		NCBI_folder = os.path.abspath(options.database) + '/NCBI'
		database_generator.NCBI_DB(data2download, NCBI_folder, Debug)

	else:
		print ("+ No update of the database has been requested using option --fast")
		
	print ("\n*************** Finish *******************")
	start_time_partial = functions.timestamp(start_time_total)

	print ("+ Exiting identification module.")
	return()

####################################
def KMA_ident(options, pd_samples_retrieved, outdir_dict, retrieve_databases, time_partial):
	"""Kmer identification using software KMA_.
	
	:param options: options passed to the :func:`BacterialTyper.modules.ident.run_ident` main function (threads, KMA_cutoff, etc). See details in...
	:param pd_samples_retrieved: pandas dataframe for samples to process.
	:param outdir_dict: dictionary containing information for each sample of the output folder for this process.
	:param retrieve_databases: 
	:param time_partial: timestamp of start time of the process.
	
	:type options: 
	:type pd_samples_retrieved: pandas.DataFrame()
	:type outdir_dict: Dictionary
	:type retrieve_databases: pandas.DataFrame()
	:type time_partial: 
	
	:return: Information of the identification. See example below.
	:rtype: pandas.DataFrame()
	
	See example of returned dataframe in file :file:`/devel/results/KMA_ident_example.csv` here:
	
	.. include:: ../../devel/results/KMA_ident_example.csv
		:literal:
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.config.set_config.get_exe`
	
		- :func:`BacterialTyper.scripts.functions.boxymcboxface`
		
		- :func:`BacterialTyper.modules.ident.send_kma_job`
		
		- :func:`BacterialTyper.modules.ident.get_outfile`
	
		- :func:`BacterialTyper.scripts.species_identification_KMA.check_db_indexed`
	
		- :func:`BacterialTyper.scripts.species_identification_KMA.parse_kma_results`
	
		
	.. include:: ../../links.inc	
	
	"""
	
	### print header
	functions.boxymcboxface("KMA Identification")

	## set defaults
	kma_bin = set_config.get_exe("kma")	

	## check status
	databases2use = []
	for	index, db2use in retrieve_databases.iterrows():
		## index_name
		if (str(db2use['source']).startswith('KMA')):
			print ('+ Check database: ' + db2use['db'])
			fold_name = os.path.dirname(db2use['path'])
			
			index_status = species_identification_KMA.check_db_indexed(db2use['path'], fold_name )
			if (index_status == True):
				print (colored("\t+ Databases %s seems to be fine...\n\n" % db2use['db'], 'green'))
				databases2use.append(db2use['path'])
			else:
				#databases2use.remove(db2use)
				print (colored("\t**Databases %s is not correctly indexed. Not using it...\n" % db2use['db'], 'red'))

	## debug message
	if (Debug):
		print (colored("**DEBUG: databases2use\n" +  "\n".join(databases2use) + "\n**", 'yellow'))

	## Start identification of samples
	print ("\n+ Send KMA identification jobs...")

	## optimize threads
	name_list = set(pd_samples_retrieved["name"].tolist())
	threads_job = functions.optimize_threads(options.threads, len(name_list)) ## threads optimization
	max_workers_int = int(options.threads/threads_job)

	## debug message
	if (Debug):
		print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
		print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
		print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))

	# Group dataframe by sample name
	sample_frame = pd_samples_retrieved.groupby(["name"])
	
	## send for each sample
	with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor:
		for db2use in databases2use:
			
			## load database on memory
			print ("+ Loading database on memory for faster identification.")
			return_code_load = species_identification_KMA.load_db(kma_bin, db2use)
			## send for each sample
			commandsSent = { executor.submit(send_kma_job, outdir_dict[name], sorted(cluster["sample"].tolist()), name, db2use, threads_job, Debug): name for name, cluster in sample_frame }

			for cmd2 in concurrent.futures.as_completed(commandsSent):
				details = commandsSent[cmd2]
				try:
					data = cmd2.result()
				except Exception as exc:
					print ('***ERROR:')
					print (cmd2)
					print('%r generated an exception: %s' % (details, exc))

			## remove database from memory
			print ("+ Removing database from memory...")
			return_code_rm = species_identification_KMA.remove_db(kma_bin, db2use)
			
			if (return_code_rm == 'FAIL'):
				print (colored("***ERROR: Removing database from memory failed. Please do it manually! Execute command: %s" %cmd_rm_db,'red'))
		
			## functions.timestamp
			time_partial = functions.timestamp(time_partial)
		
	## parse results		
	print ("+ KMA identification call finished for all samples...")
	print ("+ Parse results now")
	results_summary = pd.DataFrame()
	for db2use in databases2use:
		### [TODO]: parse data according to database: bacteria, plasmids or user data or genbank data provided
		
		basename_db = os.path.basename(db2use)
		pd.set_option('display.max_colwidth', None)
		pd.set_option('display.max_columns', None)

		###
		for name, cluster in sample_frame:
			
			## get result
			## outdir_KMA
			outdir_dict_kma = functions.create_subfolder("kma", outdir_dict[name])
			result = get_outfile(outdir_dict_kma, name, db2use)
			#print ('\t- File: ' + result + '.spa')
			
			## get results using a cutoff value [Defaulta: 80]
			results = species_identification_KMA.parse_kma_results(result + '.spa', options.KMA_cutoff)
			results['Database'] = basename_db

			### check if db2use is plasmids as it could be several.
			if (results.index.size > 1):
				if (basename_db == "plasmids.T" or basename_db == "viral.TG"):
					## let it be several entries
					results['Sample'] = name
					results_summary = results_summary.append(results, ignore_index=True)
				else:
					print (colored("###########################################", 'yellow'))
					print (colored("Sample %s contains multiple strains." %name, 'yellow'))
					print (colored("###########################################", 'yellow'))
					print (colored(results, 'yellow'))
					print ('\n\n')
					
					## add both strains if detected	
					results['Sample'] = name
					results_summary = results_summary.append(results, ignore_index=True)
					
					## TODO: add multi-isolate flag
		
			elif (results.index.size == 1): ## 1 clear reference
				results['Sample'] = name
				results_summary = results_summary.append(results, ignore_index=True)
		
			else:
				print (colored('\tNo clear strain from database %s has been assigned to sample %s' %(basename_db, name), 'yellow'))
				## add empty line if no available
				results['Sample'] = name
				results_summary = results_summary.append(results, ignore_index=True)
	
	print ("+ Finish this step...")
	
	## debug message
	if (Debug):
		results_summary.to_csv(quotechar='"')
	
	return (results_summary)

###################################
def send_kma_job(outdir_file, list_files, name, database, threads, Debug):
	"""
	Executes KMA identification jobs
	
	This function automates the process of checking if any previous run succeeded or
	runs the appropiate identification process for the sample and database provided.
	
	:param outdir_file:
	:param list_files:
	:param name:
	:param database:
	:param threads:
	:param dataFrame_sample:
	
	:type outdir_file:
	:type list_files:
	:type name:
	:type database:
	:type threads:
	:type dataFrame_sample:
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.config.set_config.get_exe`
	
		- :func:`BacterialTyper.scripts.species_identification_KMA.kma_ident_call`
	
		- :func:`BacterialTyper.module.ident.get_outfile`
		
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`
		
		
	"""
	
	if (Debug):
		print (colored("**DEBUG: ident.send_kma_job call**", 'yellow'))
		print ("outdir_file")
		print (outdir_file)
		print ("list_files")
		print (list_files)
		print ("name: " + name)
		print ("database: " + database)
		
	## outdir_KMA
	outdir_dict_kma = functions.create_subfolder("kma", outdir_file)

	## set defaults
	kma_bin = set_config.get_exe("kma")

	## get outfile
	outfile = get_outfile(outdir_dict_kma, name, database)

	## check if previously run and succeeded
	basename_tag = os.path.basename(outfile)
	filename_stamp = outdir_dict_kma + '/.success_' + basename_tag
	
	if (Debug):
		print ("Outdir: ", outdir_dict_kma)
		print ("outfile: ", outfile)
		print ("Filename_stamp: ", filename_stamp)
	
	if os.path.isfile(filename_stamp):
		stamp =	functions.read_time_stamp(filename_stamp)
		print (colored("\tA previous command generated results on: %s [%s]" %(stamp, name), 'yellow'))
	else:
		## debug message
		if (Debug):
			print (colored("**DEBUG: species_identification_KMA.kma_ident_module call**", 'yellow'))
			print ("outfile = get_outfile(outdir_dict_kma, name, db2use)")
			print ("outfile: ", outfile)
			print ("species_identification_KMA.kma_ident_module(outfile, list_files, name, database, threads) ")
			print ("species_identification_KMA.kma_ident_module" + "\t" + outfile + "\t" + str(list_files) + "\t" + name + "\t" + database + "\t" + str(threads) + "\n") 
	
		## Sparse or not
		#if any(name in basename_tag for name in ['userData_KMA', 'genbank_KMA']):
		if (basename_tag == 'userData_KMA'):
			option = ''
		else:
			option = '-Sparse '
	
		## Add option to retrieve databse from memory
		option = option + '-shm 1'
	
		# Call KMA
		species_identification_KMA.kma_ident_call(outfile, list_files, name, database, kma_bin, option, threads)
		stamp =	functions.print_time_stamp(filename_stamp)

####################################
def get_outfile(output_dir, name, index_name):
	"""
	Generates the name for the output file created
	
	:param output_dir: Absolute path to results folder 
	:param name: Name of the sample
	:param index_name: Name of the database
	
	:type output_dir: string
	:type name: string
	:type index_name: string
	
	:retruns: Output file absolute path
	"""
	basename_tag = os.path.basename(index_name)
	if Project:
		output_path = output_dir
	else:
		output_path = functions.create_subfolder(name, output_dir)
		
	out_file = output_path + '/' + name + '_' + basename_tag	
	return(out_file)
	

####################################
def edirect_ident(dataFrame, outdir_dict, Debug):
	"""Connect to NCBI for information retrieval
	
	This functions uses the software edirect_ to connect to NCBI and retrieve some information regarding samples, assemblies, publications, etc.
	
	:param dataFrame: pandas dataframe for samples to process. Result from :func:`BacterialTyper.modules.ident.KMA_ident`.
	:param outdir_dict: dictionary containing information for each sample of the output folder for this process.
	
	:type dataFrame: pandas.DataFrame()
	:type outdir_dict: Dictionary
	
	:return: Information of the identification 
	:rtype: pandas.DataFrame()
	
	See example of returned dataframe in file :file:`/devel/results/edirect_download_results.csv` here:
	
	.. include:: ../../devel/results/edirect_download_results.csv
		:literal:
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.scripts.functions.get_info_file`
		
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`
	
		- :func:`BacterialTyper.scripts.functions.print_time_stamp`

		- :func:`BacterialTyper.scripts.functions.optimize_threads`
	
		- :func:`BacterialTyper.scripts.functions.create_subfolder`
	
		- :func:`BacterialTyper.scripts.functions.boxymcboxface`
		
		- :func:`BacterialTyper.scripts.functions.is_non_zero_file`
	
		- :func:`BacterialTyper.scripts.edirect_caller.generate_docsum_call`
		
		- :func:`BacterialTyper.scripts.edirect_caller.generate_xtract_call`
		
	.. include:: ../../links.inc	
	"""
	################################################
	## TODO: What to do if multi-isolate sample?
	################################################
	
	## edirect	
	functions.boxymcboxface("EDirect information")
	print ("+ Connect to NCBI to get information from samples identified...")

	## create dataframe to return results
	edirect_frame = pd.DataFrame(columns=("sample", "genus", "species", "strain", "BioSample", "genome", "Plasmids"))

	## debugging messages
	if Debug:
		print ("*******************************************************")
		print ("Dataframe sample_results: ")
		
	# Group dataframe sample name
	sample_results = dataFrame.groupby(["Sample"])
	
	for name, grouped in sample_results:
		## debugging messages
		if Debug:
			print ("Name: ", name)
			print (grouped)
		
		## use edirect to get Species_name and entry for later identification
		edirect_folder = functions.create_subfolder('edirect', outdir_dict[name])
		
		## chromosome match
		if (len(grouped.loc[grouped['Database'] == 'bacteria.ATG']['#Template']) == 0):
			if Debug:
				print ("Name: ", name)
				print ("No chromosome match identified by kmer")
			
			genus = ''
			species = ''
			BioSample_name = ''
			AssemblyAcc = ''			
		
		else:
			nucc_entry = grouped.loc[grouped['Database'] == 'bacteria.ATG']['#Template'].values[0].split()
			## e.g. NZ_CP029680.1 Staphylococcus aureus strain AR_0215 chromosome, complete genome

			##
			out_docsum_file = edirect_folder + '/nuccore_docsum.txt'
			tmp_species_outfile = edirect_folder + '/info.csv'
			filename_stamp = edirect_folder + '/.success_species'
					
			if os.path.isfile(filename_stamp):
				stamp =	functions.read_time_stamp(filename_stamp)
				print (colored("\tA previous command generated results on: %s [%s]" %(stamp, name), 'yellow'))
				status=True	
			else: 
				edirect_caller.generate_docsum_call('nuccore', nucc_entry[0], out_docsum_file)
				status = edirect_caller.generate_xtract_call(out_docsum_file, 'DocumentSummary', 'Organism,BioSample,AssemblyAcc,Strain', tmp_species_outfile)
				
			########################################	
			## get information from edirect call
			########################################	
			if not status:
				print ("NO INFORMATION")
				continue

			taxa_name_tmp = functions.get_info_file(tmp_species_outfile)
			Organism = taxa_name_tmp[0].split(',')[0].split()
			genus = Organism[0] 							## genus
			species = Organism[1] 							## species
			BioSample_name = taxa_name_tmp[0].split(',')[1]	## BioSample
			AssemblyAcc = taxa_name_tmp[0].split(',')[2] 	## AssemblyAcc
			
			## sometimes strain is missing
			if len(taxa_name_tmp[0].split(',')) > 3:
				strain = taxa_name_tmp[0].split(',')[3] 	## strain
			else:
				strain = 'NaN'
			
			## get GenBank accession ID
			out_docsum_file_assembly = edirect_folder + '/assembly_docsum.txt'
			AssemblyAcc_outfile = edirect_folder + '/AssemblyAcc.csv'
			
			edirect_caller.generate_docsum_call('assembly', AssemblyAcc, out_docsum_file_assembly)
			edirect_caller.generate_xtract_call(out_docsum_file_assembly, 'DocumentSummary', 'Genbank', AssemblyAcc_outfile) 
			
			## some error occurred
			if not functions.is_non_zero_file(out_docsum_file_assembly):
				continue
			
			## Is it better to download Refseq or Genbank?
			## https://www.quora.com/What-is-the-difference-between-Refseq-and-Genbank		
			
			GenbankAcc = functions.get_info_file(AssemblyAcc_outfile)
			if Debug:
				print("Sample: ", name)
				print("Genbank Acc: ", GenbankAcc[0])
		
		## plasmid match
		group_plasmid = grouped.loc[grouped['Database'] == 'plasmids.T' ]
		plasmid_entries = group_plasmid['#Template'].tolist()
			## e.g. NZ_CP029083.1 Staphylococcus aureus strain AR464 plasmid unnamed1, complete sequence
		plasmid_entries_str = ",".join([i.split()[0] for i in plasmid_entries])

		## save edirect_frame
		#("sample", "taxa", strain, genome "BioSample", "Plasmids"))
		edirect_frame.loc[len(edirect_frame)] = (name, genus, species, strain, BioSample_name, GenbankAcc[0], plasmid_entries_str)

		stamp =	functions.print_time_stamp(filename_stamp)

	## debugging messages
	if Debug:
		print ("*******************************************************")
	
	return (edirect_frame)

####################################
def MLST_ident(options, dataFrame, outdir_dict, dataFrame_edirect, retrieve_databases):
	"""Generate MLST profile identification
	
	This functions uses the `MLSTar software`_ to retrieve Multi locus sequence typing (MLST) profiles from PubMLST_ for the given species previously identified by KMA. It generates MLST profiling for each sample. 
	
	:param options: options passed to the :func:`BacterialTyper.modules.ident.run_ident` main function (threads, KMA_cutoff, etc). See details in...
	:param dataFrame: pandas dataframe for samples to process. Result from :func:`BacterialTyper.modules.ident.KMA_ident`.
	:param outdir_dict: dictionary containing information for each sample of the output folder for this process.
	:param dataFrame_edirect: pandas dataframe resulted from :func:`BacterialTyper.modules.ident.edirect_ident`.
	:param retrieve_databases: 
	
	:type options: 
	:type dataFrame: pandas.DataFrame()
	:type outdir_dict: Dictionary
	:type dataFrame_edirect: pandas.DataFrame()
	:type retrieve_databases: pandas.DataFrame()
	
	:return: Information of the MLST identification. Dictionary keys are samples and values are the absolute path to file generate by :func:`BacterialTyper.scripts.MLSTar.run_doMLST` containing MLST information.
	:rtype: Dictionary

	
	See example of returned dataframe in file :file:`/devel/results/doMLST_result_example.csv` here:
	
	.. include:: ../../devel/results/doMLST_result_example.csv
		:literal:
	
	.. seealso:: Additional information to PubMLST available datasets.
	
		- :doc:`PubMLST datasets<../../../data/PubMLST_datasets>`
	
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`
	
		- :func:`BacterialTyper.scripts.functions.create_subfolder`
		
		- :func:`BacterialTyper.scripts.functions.boxymcboxface`
		
		- :func:`BacterialTyper.scripts.MLSTar.run_MLSTar`
		
		- :func:`BacterialTyper.scripts.sampleParser.get_files`
		
		- :func:`BacterialTyper.scripts.MLSTar.get_MLSTar_species`
		
	.. include:: ../../links.inc	
	"""

	## TODO: Samples might not be assembled...to take into account and return 0
	## TODO: Fix and install MLSTar during installation
	########################################################################################
	
	rscript = set_config.get_exe("Rscript") 
	########################################################################################

	## TODO: What to do if multi-isolate sample?
	## TODO: Control if a different profile is provided via --MLST_profile
	## TODO: Check time passed and download again if >?? days passed]
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: dataFrame_edirect identified**", 'yellow'))
		print (dataFrame_edirect)

	## MLST call	
	functions.boxymcboxface("MLST typing")
	print ("+ Create classical MLST typification of each sample according to species retrieved by kmer...")

	## get assembly files
	input_dir = os.path.abspath(options.input)
	assembly_samples_retrieved = sampleParser.get_files(options, input_dir, "assembly", ["fna"])

	## debug message
	if (Debug):
		print (colored("**DEBUG: assembly_samples_retrieved**", 'yellow'))
		print (assembly_samples_retrieved)	
	
	# init
	MLST_results = {}
		
	## get MLST_profile: default or provided
	mlst_profile_list = retrieve_databases.loc[ retrieve_databases['db'] == 'PubMLST']['path'].tolist()

	if (Debug):
		print ("** Debug **")
		print ("mlst_profile_list")
		print (mlst_profile_list)

		print ("dataFrame_edirect")
		print (dataFrame_edirect)


	## Generate MLST call according to species identified for each sample
	for index, row in dataFrame_edirect.iterrows():
		MLSTar_taxa_name = MLSTar.get_MLSTar_species(row['genus'], row['species'] )
		
		if (MLSTar_taxa_name == 'NaN'):
			print (colored("\t- Not available PubMLST profile for sample [%s] identified as %s %s" %(row['sample'], row['genus'], row['species']), 'yellow'))
		
		else:
			for mlst_profile in mlst_profile_list:

				## species folder
				#species_mlst_folder = functions.create_subfolder(MLSTar_taxa_name, pubmlst_folder)
				species_mlst = mlst_profile.split(',')[0]
				species_mlst_folder = mlst_profile.split(',')[1]
			
				## output file
				output_file = species_mlst_folder + '/PubMLST_available_scheme.csv'
				filename_stamp = species_mlst_folder + '/.success_scheme'
			
				## 
				if MLSTar_taxa_name == species_mlst:		
					if os.path.isfile(filename_stamp):
						stamp =	functions.read_time_stamp(filename_stamp)
						print (colored("\tA previous command generated results on: %s" %stamp, 'yellow'))
					else:
						### get scheme available
						MLSTar.getPUBMLST(MLSTar_taxa_name, rscript, output_file)
						stamp =	functions.print_time_stamp(filename_stamp)

					## parse and get scheme for classical MLST
					schemes_MLST = pd.read_csv(output_file, sep=',', header=0)

					##
					for item, cluster in schemes_MLST.iterrows():
						if cluster['len'] < 10:
							scheme2use = int(cluster['scheme'])
							continue			
					### 
					sample = row['sample']
					MLSTar_folder = functions.create_subfolder('MLST', outdir_dict[sample])
					genome_file = assembly_samples_retrieved.loc[assembly_samples_retrieved['name'] == sample]['sample'].values[0]
	
					## call MLST
					(results, profile_folder) = MLSTar.run_MLSTar(species_mlst_folder, rscript, MLSTar_taxa_name, scheme2use, sample, MLSTar_folder, genome_file, options.threads)
					MLST_results[sample] = results

	##	
	print ("+ Finish this step...")
	return (MLST_results)

####################################
def get_external_kma(kma_external_files, Debug):
	print ('\t- Get additional kma databases:')
	## external sequences provided are indexed and generated in the same folder provided 
	
	option_db = ""
	if (kma_external_files):
		kma_external_files = set(kma_external_files)		
		kma_external_files = [os.path.abspath(f) for f in kma_external_files]	
		
		## check if indexed and/or index if necessary
		external_kma_dbs_list = []
		
		## set defaults
		kma_bin = set_config.get_exe("kma")
		for f in kma_external_files:
			file_name = os.path.basename(f)
			fold_name = os.path.dirname(f)
			print (colored('\t\t+ %s' %file_name, 'green'))
			print ()

			## generate db
			databaseKMA = species_identification_KMA.generate_db([f], file_name, fold_name, 'new', 'single', Debug, kma_bin)
			if not databaseKMA:
				print (colored("***ERROR: Database provided is not indexed.\n" %databaseKMA,'orange'))
			else:
				external_kma_dbs_list.append(databaseKMA)
			
		external_kma_dbs_string = ','.join(external_kma_dbs_list)
		option_db = "kma_external:" + external_kma_dbs_string

	else:
		## rise error & exit
		print (colored("***ERROR: No database provided via --kma_external_file option.\n",'red'))
		exit()

	return(option_db)				

####################################
def get_options_db(options):
	"""Select databases to use according to the input options.
	
	:param options:
	
	:returns: Dataframe with database information among all databases available.
	"""
	
	print ("\n\n+ Select databases to use for identification:")
	
	### database folder to use
	database2use = os.path.abspath(options.database)
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: Database to use: " +  database2use + " **", 'yellow'))
	
	## according to user input: select databases to use
	option_db = ""
	
	############################################################
	## Default db KMA
	############################################################
	kma_dbs = []
	if not options.only_kma_db: ## exclusive
		kma_dbs = ["bacteria", "plasmids"]
	
	if (options.kma_dbs):
		options.kma_dbs = options.kma_dbs + kma_dbs
		options.kma_dbs = set(options.kma_dbs)		
	else:
		options.kma_dbs = kma_dbs

	## rise error & exit if no dbs provided
	if not (options.kma_dbs):
		print (colored("***ERROR: No database provided via --kma_db option.\n",'red'))
		exit()

	############################################################
	### Options:
	
	############
	## 1) only user data: previously identified and added
	############
	if (options.only_user_data):
		option_db = "user_data"
		
	############
	## 2) only genbank data: previously download from NCBI reference genomes
	############
	elif (options.only_genbank_data):
		option_db = "genbank"
	
	############
	## 3) only external kma
	############
	elif (options.only_external_kma):
		option_db = get_external_kma(options.kma_external_files, Debug)
		## rise attention
		if (options.kma_dbs):
			print (colored("***ATTENTION:\nDefatult databases and databases provided via --kma_dbs option would not be used as --only_external_kma option provided.\n",'red'))

	#################
	## all databases 
	#################
	else:		
		####################
		## default KMA dbs
		####################
		print ('\t- Selecting kma databases:')
		kma_dbs_string = ','.join(options.kma_dbs)
		option_db = "kma:" + kma_dbs_string
	
		for i in options.kma_dbs:
			print (colored('\t\t+ %s' %i, 'green'))
		
		#################
		## External file
		#################
		if (options.kma_external_files):
			option_db_tmp = get_external_kma(options.kma_external_files, Debug)
			option_db = option_db + '#' + option_db_tmp
			
		#############################
		## Previously identified data
		#############################
		if any([options.user_data, options.all_data]):
			option_db = option_db + '#kma_user_data:user_data'
	
		#############################
		## Genbank reference data
		#############################
		if any([options.genbank_data, options.all_data]):
			option_db = option_db + '#kma_NCBI:genbank'

	###############
	### PubMLST ###
	###############
	print ("\n\t - Select MLST profiles")	
	option_db_PubMLST = 'MLST:PubMLST'
	print (colored("\t\t + Default MLST profile under database provided: PubMLST", 'green'))

	if options.MLST_profile:
		## user provides a PubMLST profile
		options.MLST_profile = os.path.abspath(options.MLST_profile)
		option_db_PubMLST = option_db_PubMLST + '#MLST:'	+ options.MLST_profile
		print (colored("\t\t + User provided MLST profile: %s" %options.MLST_profile, 'green'))
	
	###############
	### get dbs
	###############
	print ("\n+ Parsing information to retrieve databases")
	print ("+ Reading from database: " + database2use)
	functions.print_sepLine("-",50, False)

	###############
	## debug message
	if (Debug):
		print (colored("**DEBUG: option_db: " +  option_db + " **", 'yellow'))
		print (colored("**DEBUG: option_db_PubMLST : " +  option_db_PubMLST  + " **", 'yellow'))

	pd_KMA = database_generator.getdbs("KMA", database2use, option_db, Debug)
	pd_PubMLST = database_generator.getdbs("MLST", database2use, option_db_PubMLST, Debug)

	functions.print_sepLine("-",50, False)

	## return both dataframes
	pd_Merge = pd.concat([pd_KMA, pd_PubMLST], sort=True, ignore_index=True)
	return (pd_Merge)
